use actix_web::{web::Data, FromRequest};
use serde::{Deserialize, Serialize};

use chrono::NaiveDateTime;
use diesel::prelude::*;
use std::pin::Pin;

use crate::{schema::posts::dsl::{id as schema_post_id, posts}, State};

#[derive(Queryable, Serialize, Deserialize)]
pub struct Post {
    pub id: i32,
    pub title: String,
    pub content: String,
    pub author_id: i32,
    pub tags: Vec<Option<String>>,
    pub created_at: NaiveDateTime,
    pub updated_at: Option<NaiveDateTime>,
    pub last_updated_by: Option<i32>
}

#[derive(Insertable, Serialize, Deserialize)]
#[diesel(table_name = crate::schema::posts)]
pub struct NewPost {
    pub title: String,
    pub content: String,
    pub author_id: i32,
    pub tags: Option<Vec<String>>,
}

#[derive(Serialize, Deserialize, AsChangeset)]
#[diesel(table_name = crate::schema::posts)]
pub struct EditPost {
    pub title: String,
    pub content: String,
    pub tags: Option<Vec<String>>,
}

impl FromRequest for Post {
    type Error = actix_web::error::Error;
    type Future = Pin<Box<dyn Future<Output = Result<Self, Self::Error>>>>;

    fn from_request(
        req: &actix_web::HttpRequest,
        _: &mut actix_web::dev::Payload,
    ) -> Self::Future {
        let req = req.clone();

        Box::pin(async move {
            let state = req.app_data::<Data<State>>();
            if state.is_none() {
                return Err(actix_web::error::ErrorInternalServerError(
                    "cannot retrieve app state, therefore can not retrieve post data",
                ));
            }

            let state = state.unwrap();
            let mut conn = match state.db.get() {
                Ok(conn) => conn,
                Err(..) => Err(actix_web::error::ErrorInternalServerError(
                    "cannot retrieve database connection from the pool",
                ))?,
            };

            let post_id = actix_web::web::Path::<i32>::extract(&req).await;
            match post_id {
                Ok(id) => {
                    let post = posts
                        .filter(schema_post_id.eq(id.into_inner()))
                        .first::<Post>(&mut conn)
                        .optional();

                    match post {
                        Ok(Some(post)) => Ok(post),
                        Ok(None) => Err(actix_web::error::ErrorBadRequest(
                            "post with this ID is not found",
                        )),
                        Err(why) => Err(actix_web::error::ErrorInternalServerError(why)),
                    }
                }
                Err(why) => Err(actix_web::error::ErrorInternalServerError(why)),
            }
        })
    }
}
