use actix_web::{
    dev::Payload,
    error::{ErrorBadRequest, ErrorInternalServerError, ErrorNotFound},
    web::Data,
    Error, FromRequest, HttpRequest,
};

use chrono::Utc;
use jsonwebtoken::{DecodingKey, EncodingKey, Header, Validation};
use serde::{Deserialize, Serialize};
use std::future::{self, ready};

use crate::schema::users::dsl::{id as schema_user_id, users};
use crate::{database::models::user::User, State};
use diesel::{ExpressionMethods, QueryDsl, RunQueryDsl};

static WEEK: i64 = 60 * 60 * 24 * 7;

#[derive(Debug, Serialize, Deserialize)]
pub struct Claims {
    pub user_id: i32,
    pub exp: usize,
}

impl Claims {
    pub fn generate_token(secret: &str, user_id: i32) -> String {
        let now = Utc::now().timestamp();

        let payload = Claims {
            exp: (now + WEEK) as usize,
            user_id,
        };

        jsonwebtoken::encode(
            &Header::default(),
            &payload,
            &EncodingKey::from_secret(secret.as_ref()),
        )
        .unwrap()
    }
}

#[derive(Deserialize, Debug, Serialize)]
pub struct AuthorizedUser {
    pub token: String,
    pub user_id: i32,
    pub user: User,
}

fn get_user(state: &Data<State>, user_id: i32) -> Option<User> {
    let mut conn = match state.db.get() {
        Ok(conn) => conn,
        Err(..) => None?,
    };

    let user = users
        .filter(schema_user_id.eq(user_id))
        .first::<User>(&mut conn);

    match user {
        Ok(user) => Some(user),
        Err(..) => None,
    }
}

impl FromRequest for AuthorizedUser {
    type Error = Error;
    type Future = std::future::Ready<Result<Self, Self::Error>>;

    fn from_request(req: &HttpRequest, _: &mut Payload) -> Self::Future {
        let state = req.app_data::<Data<State>>();
        let header = req.headers().get("Authorization");
        if header.is_none() {
            return future::ready(Err(ErrorBadRequest("no Authorization header specified")));
        }

        if state.is_none() {
            return future::ready(Err(ErrorInternalServerError("cannot retrieve JWT token")));
        }

        let state = state.unwrap();
        match header.unwrap().to_str() {
            Ok(v) => {
                let parts: Vec<&str> = v.split(" ").collect();

                if parts.len() != 2 {
                    return ready(Err(ErrorBadRequest("unauthorized")));
                }

                let claims = jsonwebtoken::decode::<Claims>(
                    parts[1],
                    &DecodingKey::from_secret(state.config.jwt_secret.as_ref()),
                    &Validation::default(),
                );

                if claims.is_err() {
                    return ready(Err(ErrorBadRequest("unauthorized")));
                }

                let claims = claims.unwrap().claims;
                let user = get_user(state, claims.user_id);

                match user {
                    Some(user) => {
                        let authorized_user = AuthorizedUser {
                            user_id: claims.user_id,
                            token: v.to_string(),
                            user,
                        };

                        ready(Ok(authorized_user))
                    }
                    None => ready(Err(ErrorNotFound(
                        "invalid login credentials, user not found",
                    ))),
                }
            }
            Err(e) => ready(Err(ErrorBadRequest(e))),
        }
    }
}
