use actix_web::{
    get,
    http::StatusCode,
    patch, put,
    web::{scope, Data, Json, ServiceConfig},
    HttpResponse, Responder,
};
use diesel::{insert_into, ExpressionMethods, QueryDsl, RunQueryDsl};

use crate::utils::response::failure;
use crate::State;
use crate::{database::models::post::EditPost, schema::posts::dsl as posts_dsl};
use crate::{
    database::models::post::{NewPost, Post},
    utils::jwt::AuthorizedUser,
};

#[get("")]
async fn get(state: Data<State>) -> impl Responder {
    let mut conn = match state.db.get() {
        Ok(conn) => conn,
        Err(why) => {
            return HttpResponse::InternalServerError()
                .body(format!("database connection error: {}", why))
        }
    };

    match posts_dsl::posts.load::<Post>(&mut conn) {
        Ok(data) => HttpResponse::Ok().json(data),
        Err(why) => HttpResponse::InternalServerError().body(why.to_string()),
    }
}

#[put("")]
async fn create_post(
    state: Data<State>,
    auth: AuthorizedUser,
    post: Json<NewPost>,
) -> impl Responder {
    let mut conn = match state.db.get() {
        Ok(conn) => conn,
        Err(why) => {
            return HttpResponse::InternalServerError()
                .body(format!("database connection error: {}", why))
        }
    };

    match insert_into(posts_dsl::posts)
        .values((post.0, posts_dsl::author_id.eq(auth.user_id)))
        .execute(&mut conn)
    {
        Ok(_) => HttpResponse::new(StatusCode::NO_CONTENT),
        Err(why) => HttpResponse::InternalServerError().json(failure(why.to_string())),
    }
}

#[patch("/{id}")]
async fn update_post(
    state: Data<State>,
    auth: AuthorizedUser,
    post: Post,
    new_post: Json<EditPost>,
) -> impl Responder {
    let mut conn = match state.db.get() {
        Ok(conn) => conn,
        Err(why) => {
            return HttpResponse::InternalServerError()
                .body(format!("database connection error: {}", why))
        }
    };

    // TODO: Security, anti-abuse checks
    match diesel::update(posts_dsl::posts.filter(posts_dsl::id.eq(post.id)))
        .set((
            new_post.0,
            posts_dsl::updated_at.eq(diesel::dsl::now),
            posts_dsl::last_updated_by.eq(auth.user_id),
        ))
        .execute(&mut conn)
    {
        Ok(_) => HttpResponse::new(StatusCode::NO_CONTENT),
        Err(why) => HttpResponse::InternalServerError().json(failure(why.to_string())),
    }
}

pub fn init_routes(cfg: &mut ServiceConfig) {
    cfg.service(scope("/posts").service(get).service(create_post));
}
